/*====================================================================================================*
* PINEWOOD DERBY - PWD 4-Lane Track Timer                                  Version 5.0 - 03 June 2021 *
* Created by: Bill van Leeuwen (email: Billv923@outlook.com) for the Arduino UNO microcontroller      *
* Version 5.0 has the following features:                                                             *
*   1. Lane masking capability                                                                        *
*   2. Time accuracy of 4 decimal places (0.0001 seconds).                                            *
*   3. Ability to send the Manual Reset command to the PWD Track Timer display software               *
*   4. COM check acknowledge                                                                          *
*                                                                                                     *
* This program is designed to work in conjunction with the author's PWD race management software      *
* PWD_RaceManager04.pde which runs on a PC attached to the Arduino via the USB serial port.           *
*                                                                                                     *
* This program measures the start-to-finish time of 4 lanes. Race status and finish times are sent to * 
* the serial port to interface with the PWD race management/display software running on the PC. Time  *
* is measured from activation of the start switch to activation of the finish line sensor of each     *
* lane. Each car has 10 seconds to finish the race. If a car doesn't cross the finish line within 10  *
* seconds, it is assumed that the car had a problem or the lane was not used. In either case, the     *
* value 9.9999 is assigned to the lane time as an error indication for use by the PWD race management *
* display software running on the PC.                                                                 *
*                                                                                                     *
* This design uses optical sensors (one for each lane) embedded in the track at the finish line to    *
* detect when a car crosses the finish line. Incandescent lamps or other infrared light source        *
* mounted above the finish line sensors provide the illumination needed by the sensors. A switch      *
* connected to the start gate is used to start the timer. Lane times and track status codes are sent  * 
* via the USB port to the PWD race management/display software running on the PC.                     *
*                                                                                                     *
* Refer to the Schematic Diagram, parts list and design notes that accompany this software for wiring * 
* of the infrared lane sensors, race start switch, and reset switch.                                  *
* Feel free to contact me at Billv923@outlook.com if any questions.                                   *
*====================================================================================================*/

//Initialization
#define Num_Lanes    4         //Number of lanes
#define NotReady     0         //Race State = Not Ready
#define Ready        1         //Race State = Ready
#define Racing       2         //Race State = Racing
#define Finished     3         //Race State = Finished
#define Msg_Reset   "R"        //Reset character received from PC via Serial port
#define Msg_Comchk  "C"        //COM Check request received from PC via Serial port
#define Msg_L1Mask  "M1"       //Mask lane 1 request received from PC via Serial port
#define Msg_L2Mask  "M2"       //Mask lane 2 request received from PC via Serial port
#define Msg_L3Mask  "M3"       //Mask lane 3 request received from PC via Serial port
#define Msg_L4Mask  "M4"       //Mask lane 4 request received from PC via Serial port
#define Msg_UnMask  "U"        //Unmask all lanes request received from PC via Serial port

//Pin Assignments
#define fs1          2         //Finish Sensor Lane 1 - Digital Input Pin 2
#define fs2          3         //Finish Sensor Lane 2 - Digital Input Pin 3
#define fs3          4         //Finish Sensor Lane 3 - Digital Input Pin 4
#define fs4          5         //Finish Sensor Lane 4 - Digital Input Pin 5
#define gate_sw      8         //Gate Switch (Start)  - Digital Input Pin 8
#define reset_sw     9         //Manual Reset         - Digital Input Pin 9
#define LED         13         //IO Pin 13 LED        - Digital Pin 13


//Variable Initialization
int i;                         //Working variable
int lane_place[Num_Lanes];     //Lane finish place array
boolean sens_ready;            //False if any lane sensor is obstructed
boolean track_ready;           //False if any lane sensor is obstructed or Gate Switch Open
boolean Lane1_Mask;            //True if lane 1 is masked
boolean Lane2_Mask;            //True if lane 2 is masked
boolean Lane3_Mask;            //True if lane 3 is masked
boolean Lane4_Mask;            //True if lane 4 is masked
int Active_Lanes;              //Number of active lanes (Total lanes - masked lanes)
int Masked_Lanes;              //Number of lanes that are masked
int lane;                      //Lane number
int finish_order;              //Finish order
int racestate;                 //0=NotReady, 1=Ready, 2=Racing, 3=Finished
int fincount;                  //Tallies cars crossing finish line to determine race over
unsigned long Start_Time;      //Time (microseconds)
unsigned long Exp_Time;        //Expiration Time (milliseconds)
unsigned long Clk_Time;        //CPU Clock Time (microseconds)
float Time;                    //Working variable - Calculated Lane Time (seconds)
float lane_time[Num_Lanes];    //Calculated lane time data array (seconds)
char SMsg0[4] = "NRD";         //Serial message - Not Ready
char SMsg1[4] = "RDY";         //Serial message - Ready
char SMsg2[4] = "RAC";         //Serial message - Racing
char SMsg3[4] = "FIN";         //Serial message - Finished
char SMsg4[4] = "GSW";         //Serial message - Gate Switch not ready
char SMsg5[4] = "TRK";         //Serial message - Track Sensor(s) blocked
char SMsg6[4] = "RST";         //Serial message - Manual RESET executed
String serial_data;            //Serial data string received from PC
String bad_sens = "";          //Message string populated with obstructed lane sensors

//======================================= SETUP ROUTINE =======================================
void setup() {  
//Hardware Setup
  //Initialize Finish Sensor, Gate Switch and Reset Switch pins as inputs & pin 13 as an output
  pinMode(fs1,INPUT_PULLUP);          //Note: Sensor inputs fs1 thru fs4 transition from LOW to
  pinMode(fs2,INPUT_PULLUP);          //      HIGH when light beam is interupted (e.g. car
  pinMode(fs3,INPUT_PULLUP);          //      crosses finish line)
  pinMode(fs4,INPUT_PULLUP);
  pinMode(gate_sw,INPUT_PULLUP);      //Gate_sw is normally LOW & transitions to HIGH at race start
  pinMode(reset_sw,INPUT_PULLUP);     //Reset_sw is normally HIGH & transitions to LOW when pressed
  pinMode(LED, OUTPUT);               //set so we can turn off the pin 13 LED
  
//Software Setup
  Serial.begin(9600);                 //Enable serial comms to PC at 9600 baud
  racestate = NotReady;               //Set initial race state to not ready
  Lane1_Mask = false;                 //Preset Lane 1 mask flag to false
  Lane2_Mask = false;                 //Preset Lane 2 mask flag to false
  Lane3_Mask = false;                 //Preset Lane 3 mask flag to false
  Lane4_Mask = false;                 //Preset Lane 4 mask flag to false
  Masked_Lanes = 0;                   //Preset masked lane count to zero
  digitalWrite(LED,LOW);              //Turn off Pin 13 LED just for the heck of it
}

//======================================== MAIN LOOP ============================================
void loop() {
  //READ & PROCESS INCOMING SERIAL DATA
  serial_data = get_serial_data();                //Get any received serial data
  if (serial_data.indexOf(Msg_Reset)>-1) {        //Check for Reset request from PC
    Check_Track();                                //If so, call Check Track subroutine
    if (track_ready){                             //and only do a reset if Track is ready
      initialize();                               //Call initialize function & reset everything
      racestate = Ready;                          //Set racestate to Ready
      Serial.println(SMsg1);                      //Send Ready state message to PC
    } 
    else {                                        //Do this if track is not ready
      racestate = NotReady;                       //Set racestate to Not Ready
      Serial.println(SMsg0);                      //Send Not Ready state message to PC
    }
  }
  if (serial_data.indexOf(Msg_UnMask)>-1) {       //Check for lane Unmask request from GPRM
    Masked_Lanes = 0;                             //Clear masked lane count
    Lane1_Mask = false;                           //Clear Lane 1 mask flag
    Lane2_Mask = false;                           //Clear Lane 2 mask flag
    Lane3_Mask = false;                           //Clear Lane 3 mask flag
    Lane4_Mask = false;                           //Clear Lane 4 mask flag
  }
  if (serial_data.indexOf(Msg_L1Mask)>-1) {       //Check for lane 1 mask request from GPRM
    Lane1_Mask = true;                            //Set to TRUE if so
    ++Masked_Lanes;                               //Increment masked lane count
  }
  if (serial_data.indexOf(Msg_L2Mask)>-1) {       //Check for lane 2 mask request from GPRM
    Lane2_Mask = true;                            //Set to TRUE if so
    ++Masked_Lanes;                               //Increment masked lane count
  }
  if (serial_data.indexOf(Msg_L3Mask)>-1) {       //Check for lane 3 mask request from GPRM
    Lane3_Mask = true;                            //Set to TRUE if so
    ++Masked_Lanes;                               //Increment masked lane count
  }
  if (serial_data.indexOf(Msg_L4Mask)>-1) {       //Check for lane 4 mask request from GPRM
    Lane4_Mask = true;                            //Set to TRUE if so
    ++Masked_Lanes;                               //Increment masked lane count
  }
  if (serial_data.indexOf(Msg_Comchk)>-1) {       //Check for COM Port check request from PC
    Serial.println("@");                          //Send COM Port check acknowledge message to PC
  }
 
  //CHECK & PROCESS RACE STATES
  //Process activation of the Manual Reset switch
  if (digitalRead(reset_sw)==0){                  //Manual timer Reset button pressed?
    Serial.println(SMsg6);                        //Send Manual Reset message to PC
    delay(500);                                   //Insert delay to debounce reset switch
    Lane1_Mask = false;                           //Clear Lane 1 mask flag in case it was set
    Lane2_Mask = false;                           //Clear Lane 2 mask flag in case it was set
    Lane3_Mask = false;                           //Clear Lane 3 mask flag in case it was set
    Lane4_Mask = false;                           //Clear Lane 4 mask flag in case it was set
    Masked_Lanes = 0;                             //Clear masked lane count in case it was set
    Check_Track();                                //Call Check Track subroutine
    if (track_ready){                             //and only do a reset if Track is ready
      initialize();                               //Call initialize function & reset everything
      racestate = Ready;                          //Set racestate to Ready
      Serial.println(SMsg1);                      //Send Ready state message to PC
    } 
    else {                                        //Do this if track is not ready
      racestate = NotReady;                       //Set racestate to NotReady
      Serial.println(SMsg0);                      //Send Not Ready state message to PC
    }
  }
  // Transition from Ready to Racing when the Gate switch is tripped
  if (racestate == Ready && digitalRead(gate_sw)==1) {
    Start_Time = micros();                        //Get start time from CPU clock in microseconds
    Exp_Time = millis() + 10000;                  //Set Expiration = Start_Time + 10 seconds (milliseconds)
    racestate = Racing;                           //Set racestate to Racing
    Serial.println(SMsg2);                        //Send Racing state message to PC
    delay(200);                                   //Wait a while before we start finish check loop
    
    //Race underway. Check/Wait for everyone to finish
    /*====================================================================================================*
    * NOTE!!!!: The 'while' loop below has been optimized to repeatedly check lane sensors as fast as     *
    * possible by accessing the digital port input register directly via the command 'bitRead(PIND,fsx)'  *
    * instead of the command 'digitalRead()'. This command is specifically tailored to Arduino boards     *
    * using the ATmega8, ATmega168 or ATmega328p chips. If this code is used with an Arduino clone having *
    * a different processing chip, it may not work. If this is the case, replace each occurence of the    *
    * 'bitRead(PIND,fsx)' command with 'digitalRead(fsx)' where x is the lane number.                     *
    *====================================================================================================*/
    Active_Lanes = Num_Lanes - Masked_Lanes;      //Calculate number of active lanes
    while (fincount<Active_Lanes) {               //Loop until all cars have crossed finish line
      Clk_Time = micros();                        //Get the CPU clock time in microseconds
      if (lane_place[0]==0) {                     //Check & skip if Lane 1 already finished
        if (!Lane1_Mask) {                        //Check & skip if Lane 1 is masked
          if (bitRead(PIND, fs1)) {               //Is Car over Lane 1 Finish Line?
            lane_time[0]=Clk_Time;                //Yes...get time
            lane_place[0]=++fincount;             //Incr finish count & set finish place for Lane 1
          }
        }
      }
      if (lane_place[1]==0) {                     //Check & skip if Lane 2 already finished
        if (!Lane2_Mask) {                        //Check & skip if Lane 2 is masked
          if (bitRead(PIND, fs2)) {               //Is Car over Lane 2 Finish Line?
            lane_time[1]=Clk_Time;                //Yes...get time
            lane_place[1]=++fincount;             //Incr finish count & set finish place for Lane 2
          }
        }
      }  
      if (lane_place[2]==0) {                     //Check & skip if Lane 3 already finished
        if (!Lane3_Mask) {                        //Check & skip if Lane 3 is masked
          if (bitRead(PIND, fs3)) {               //Is Car over Lane 3 Finish Line?
            lane_time[2]=Clk_Time;                //Yes...get time
            lane_place[2]=++fincount;             //Incr finish count & set finish place for Lane 3
          }
        }
      }
      if (lane_place[3]==0) {                     //Check & skip if Lane 4 already finished
        if (!Lane4_Mask) {                        //Check & skip if Lane 4 is masked
          if (bitRead(PIND, fs4)) {               //Is Car over Lane 4 Finish Line?
            lane_time[3]=Clk_Time;                //Yes...get time
            lane_place[3]=++fincount;             //Incr finish count & set finish place for Lane 4
          }
        }
      }
      if (millis()>Exp_Time) break;               //Exit 'while' loop if time has expired
    }
    //Race is over or time has expired
    racestate=Finished;                           //Set racestate to Finished
    //Send Race Results To PC as one long string (done to simplify processing at the PC)
    Serial.print("Times:");                       //Precede string with "Times" as an identifier
    for (int n=0; n<Num_Lanes; n++) {             //Loop thru all lanes
      Serial.print(" ");                          //Add a space to separate numbers
      if (lane_time[n]>0) {                       //Check to see if lane time is valid
        Time = (float)((lane_time[n]-Start_Time)/1000000.0);  //Calc lane time and convert to seconds
        Serial.print(Time,4);                     //Print lane time trimmed to 4 places after decimal
      }
      else {
        Serial.print("9.9999");                   //Print all nines if lane time is invalid
      }
    }
    Serial.println("");                           //Send a cr/lf to indicate end of the string
    delay(300);                                   //Wait a while to allow display SW to process times
    Serial.println(SMsg3);                        //Send Finish state message to PC
    delay(100);                                   //Wait a while to allow display SW to process SMsg3
  }                                               //End Racing "if"
}                                                 //Loop Back

/*================================================================================*
  SUBROUTINES & FUNCTIONS
 *================================================================================*/
//Subroutine: Initialize Racing Variables
void initialize() {  
  for (int n=0; n<Num_Lanes; n++) {           //Reset lane Place & Times
    lane_place[n] = 0;
    lane_time[n] = 0;
  }
  fincount = 0;                               //Reset finish count to 0
  Serial.flush();                             //Clear the serial input buffer
  return;
}

//Function: Read Serial Data
String get_serial_data() {
  String Sdata = "";                          //Clear any old data
  if (Serial.available() > 0) {               //Check for new data line
    Sdata = Serial.readStringUntil('\n');     //and get it if so
  }
  return Sdata;                               //Return serial data to main program
}

//Function: Check Track - Sets 'track_ready' to true or false
int Check_Track() {
  track_ready = true;                         //Default track ready status to true
  sens_ready = true;                          //Default sensor ready status to true
  bad_sens = "";                              //Clear bad sensor string
  if (digitalRead(gate_sw)==1) {              //Check if Gate Switch is open
    Serial.println(SMsg4);                    //If so, send Gate Switch not ready message to PC
    track_ready = false;                      //And set track ready status to false
  }
  for (int n=0; n<Num_Lanes; n++) {           //Loop through all 4 lanes and
    if (digitalRead(n+2)==1) {                //Check if a lane sensor (pins 2-5) is obstructed
      sens_ready = false;                     //If so, set sensor ready status to false
      bad_sens = bad_sens + ", " + (n+1);     //Build string identifying obstructed lane sensors
    }
  }
  if (sens_ready == false) {                  //Check if a lane sensor was not ready
    Serial.println(SMsg5 + bad_sens);         //If so, send Lane Sensors not ready message to PC
    track_ready = false;                      //And set track ready status to false
  }
  delay(200);                                 //Wait a while to allow display SW to process messages
}
